//! Parse the mapping between samples ids and the corresponding module values,
//! guessing the module letter from the file.

use std::{
    collections::{hash_map::Entry, HashMap, HashSet},
    io::Read,
    num::ParseIntError,
};

use snafu::{ensure, OptionExt, ResultExt, Snafu};

use crate::{
    ful,
    io::{self, Context, ContextualizedReader}, MODULE_LETTERS,
};

#[derive(Debug)]
pub(crate) struct Samples {
    pub(crate) letter: u8,
    pub(crate) map: HashMap<usize, Vec<u8>>, // {module number for this letter ↦ sample id}
}

impl Samples {
    pub(crate) fn parse<R: Read>(mut reader: ContextualizedReader<R>) -> Result<Self, Error> {
        let parse = |reader: &mut ContextualizedReader<_>| -> Result<Self, ParseError> {
            let mut map = HashMap::new();
            let mut letter = None;
            let mut code = Vec::new();
            let mut samples = HashSet::new(); // (to check for duplicates)
            loop {
                // Extract sample id.
                let mut sample = Vec::new();
                let n_read = reader.read_until(b'\t', &mut sample)?;
                if n_read == 0 {
                    return Ok(Self {
                        map,
                        letter: letter.with_context(|| NoLetterErr)?,
                    });
                }
                sample.pop();
                ensure!(
                    samples.insert(sample.clone()),
                    DuplicateSampleErr { id: sample.clone() }
                );
                // Retrieve corresponding module code.
                let n_read = reader.read_until(b'\n', &mut code)?;
                ensure!(
                    n_read > 0,
                    EofErr {
                        what: "module code number",
                    }
                );
                if code.ends_with(b"\n") {
                    code.pop();
                }
                // Check code letter.
                let first = code[0];
                if let Some(before) = letter {
                    ensure!(before == first, TwoLettersErr { now: first, before });
                } else {
                    ensure!(
                        MODULE_LETTERS.contains(&first),
                        InvalidLetterErr { letter: first }
                    );
                    letter = Some(first);
                }
                // Check code number.
                let number: usize = String::from_utf8_lossy(&code[1..])
                    .parse()
                    .with_context(|_| NotACodeErr { code: code.clone() })?;
                ensure!((1..=96).contains(&number), InvalidCodeErr { code });
                // Record and move on to the next line.
                match map.entry(number) {
                    Entry::Occupied(o) => {
                        return DuplicateModuleErr {
                            letter: letter.unwrap(),
                            number,
                            a: ful(o.get()),
                            b: ful(&sample),
                        }
                        .fail();
                    }
                    Entry::Vacant(e) => e.insert(sample),
                };
                code.clear(); // Recycle allocation.
                reader.bump_line();
            }
        };
        parse(&mut reader).map_err(|source| Error::Parse {
            context: reader.into(),
            source,
        })
    }
}

//==================================================================================================
#[derive(Debug, Snafu)]
pub enum Error {
    #[snafu(transparent)]
    Io { source: io::Error },
    #[snafu(display("{context}{source}"))]
    Parse {
        context: Context,
        source: ParseError,
    },
}

#[derive(Debug, Snafu)]
#[snafu(context(suffix(Err)))]
pub enum ParseError {
    #[snafu(transparent)]
    Io { source: io::Error },
    #[snafu(display("Could not determine the barcode module letter from the schema: no lines?"))]
    NoLetter,
    #[snafu(display("Unexpected EOF: missing {what}."))]
    Eof { what: &'static str },
    #[snafu(display(
        "Invalid barcode module letter {letter:?}, \
         expected one of {:?}.",
         MODULE_LETTERS.map(|u| u as char),
    ))]
    InvalidLetter { letter: char },
    #[snafu(display("Not a code number: {:?}\n{source}", ful(code)))]
    NotACode {
        code: Vec<u8>,
        source: ParseIntError,
    },
    #[snafu(display("Invalid code number: {:?}", ful(code)))]
    InvalidCode { code: Vec<u8> },
    #[snafu(display(
        "The barcode module sample letter is {now:?} but previous lines were {before:?}."
    ))]
    TwoLetters { now: char, before: char },
    #[snafu(display("Sample {:?} appears twice in the schema.", ful(id)))]
    DuplicateSample { id: Vec<u8> },
    #[snafu(display(
        "Barcode module {letter}{number} is mapped to both samples {a:?} and {b:?}."
    ))]
    DuplicateModule {
        letter: char,
        number: usize,
        a: String,
        b: String,
    },
}
