//! Describe reader threads and how to handle them.
//! This thread reads from one input file, unzips, parses fastq blocks
//! and sends them to the main thread.
//! Once consumed, blocks are sent back from the main thread
//! for recycling allocations here.

use std::{
    io::Read,
    path::PathBuf,
    sync::{
        mpsc::{Receiver, SendError, SyncSender, TryRecvError},
        Arc, Mutex,
    },
    thread::JoinHandle,
};

use snafu::{ResultExt, Snafu};

use crate::{fastq::BlocksReader, io::Context, parse, Block};

// Messages sent.
type Input = Block; // Allocations to recycle.
type Output = Option<Block>; // None means correct termination.

/// Owned by the reader thread.
pub(crate) struct Inner<R: Read> {
    pub(crate) reader: BlocksReader<R>,
    pub(crate) input: Receiver<Input>,
    // I1/I2 reads are fast: use a sync channel
    // to not spoil memory with queued blocks.
    pub(crate) output: SyncSender<Output>,
    // (for runtime statistics)
    pub(crate) new_allocs: Arc<Mutex<u64>>,
    pub(crate) recycled_allocs: Arc<Mutex<u64>>,
}

/// Owned by the main thread to handle a reader thread.
pub(crate) struct Outer {
    pub(crate) infix: &'static str,
    pub(crate) path: PathBuf,
    pub(crate) handle: JoinHandle<Result<(), Error>>,
    pub(crate) output: Receiver<Output>,
    pub(crate) input: SyncSender<Input>,
}

impl<R: Read> Inner<R> {
    /// Reader thread main loop.
    pub(crate) fn run(self) -> Result<(), Error> {
        let Self {
            mut reader,
            input,
            output,
            new_allocs,
            recycled_allocs,
        } = self;
        let err = |e: SendError<Output>| Error::Message { erm: e.to_string() };
        while reader
            .parse_next_block()
            .with_context(|_| ReadErr { context: &reader })?
        {
            // Get a recycled allocation for the next block if available,
            // or create a new one.
            let block = match input.try_recv() {
                Ok(mut free) => {
                    {
                        *recycled_allocs
                            .lock()
                            .expect("poisoned recycled input allocs counter?") += 1;
                    }
                    reader.swap(&mut free);
                    free
                }
                Err(TryRecvError::Empty) => {
                    {
                        *new_allocs
                            .lock()
                            .expect("poisoned new input allocs counter?") += 1;
                    }
                    reader.block().clone()
                }
                Err(TryRecvError::Disconnected) => {
                    panic!("Main thread disconnected: reader panic.");
                }
            };
            output.send(Some(block)).map_err(err)?;
        }
        output.send(None).map_err(err)?; // Signify correct termination.
        Ok(())
    }
}

#[derive(Debug, Snafu)]
#[snafu(context(suffix(Err)))]
pub enum Error {
    #[snafu(display("{context}Could not read fastq block:\n{source}"))]
    Read {
        source: parse::fastq::Error,
        context: Context,
    },
    #[snafu(display("Message could not be sent from reader thread to main thread:\n{erm}"))]
    Message {
        erm: String, // Rendered underlying error message.
    },
}
